#include "mathext.h"

#include <stdio.h>
#include <string.h>

static float f_tolerance = 0.0001f;

vector3 vector3_new(float x, float y, float z) {
    vector3 v;

    v.x = x;
    v.y = y;
    v.z = z;

    return v;
}

void vector3_set(vector3 *v, float x, float y, float z) {
    v->x = x;
    v->y = y;
    v->z = z;
}

float vector3_magnitude(vector3 *v) {
    return (float)sqrt(v->x*v->x + v->y*v->y + v->z*v->z);
}

void vector3_normalize(vector3 *v) {
    float m = (float)sqrt(v->x*v->x + v->y*v->y + v->z*v->z);
    if (m < f_tolerance) {
        m = 1.0f;
    }
    v->x /= m;
    v->y /= m;
    v->z /= m;

    if (fabs(v->x) < f_tolerance) { v->x = 0.0f; }
    if (fabs(v->y) < f_tolerance) { v->y = 0.0f; }
    if (fabs(v->z) < f_tolerance) { v->z = 0.0f; }
}

void vector3_reverse(vector3 *v_in, vector3 *v_out) {
    v_out->x = -v_in->x;
    v_out->y = -v_in->y;
    v_out->z = -v_in->z;
}

void vector3_add(vector3 *v1_in, vector3 *v2_in, vector3 *v_out) {
    v_out->x = v1_in->x + v2_in->x;
    v_out->y = v1_in->y + v2_in->y;
    v_out->z = v1_in->z + v2_in->z;
}

void vector3_sub(vector3 *v1_in, vector3 *v2_in, vector3 *v_out) {
    v_out->x = v1_in->x - v2_in->x;
    v_out->y = v1_in->y - v2_in->y;
    v_out->z = v1_in->z - v2_in->z;
}

void vector3_mul(vector3 *v, float s) {
    v->x *= s;
    v->y *= s;
    v->z *= s;
}

void vector3_div(vector3 *v, float s) {
    v->x /= s;
    v->y /= s;
    v->z /= s;
}

void vector3_neg(vector3 *v) {
    v->x = -v->x;
    v->y = -v->y;
    v->z = -v->z;
}

void vector3_vector_product(vector3 *v1_in, vector3 *v2_in, vector3 *v_out) {
    v_out->x = v1_in->y * v2_in->z - v1_in->z * v2_in->y;
    v_out->y = -v1_in->x * v2_in->z + v1_in->z * v2_in->x;
    v_out->z = v1_in->x * v2_in->y - v1_in->y * v2_in->x;
}

float vector3_scalar_product(vector3 *v1, vector3 *v2) {
    return v1->x * v2->x + v1->y * v2->y + v1->z * v2->z;
}

float vector3_triple_scalar_product(vector3 *u, vector3 *v, vector3 *w) {
    return (u->x * (v->y * w->z - v->z * w->y)) +
            (u->y * (-v->x * w->z + v->z * w->x)) +
            (u->z * (v->x * w->y - v->y * w->x));
}

void dump_matrix4x4(float *matrix) {
    fprintf(stdout, "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\t%.2f\n",
                    matrix[0], matrix[1], matrix[2], matrix[3],
                    matrix[4], matrix[5], matrix[6], matrix[7],
                    matrix[8], matrix[9], matrix[10], matrix[11],
                    matrix[12], matrix[13], matrix[14], matrix[15]);
}

void dump_vector4(float *vector) {
    fprintf(stdout, "%.2f\n"
                    "%.2f\n"
                    "%.2f\n"
                    "%.2f\n",
                    vector[0], vector[1], vector[2], vector[3]);
}

void dump_matrix3x3(float *matrix) {
    fprintf(stdout, "%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\n"
                    "%.2f\t%.2f\t%.2f\n",
                    matrix[0], matrix[1], matrix[2],
                    matrix[3], matrix[4], matrix[5],
                    matrix[6], matrix[7], matrix[8]);
}

void dump_vector3(float *vector) {
    fprintf(stdout, "%.2f\n"
                    "%.2f\n"
                    "%.2f\n",
                    vector[0], vector[1], vector[2]);
}

void dump_coords_3d(float *coords, unsigned long coords_size) {
	printf("[\n");
	for (int i = 0; i < coords_size / sizeof(float); i += 3) {
		printf("x: %.2f,  y: %.2f,  z: %.2f\n", coords[i], coords[i+1], coords[i+2]);
	}
	printf("]\n");
}

void mat3x3_from_mat4x4(float out_mat3x3[9], float in_mat4x4[16]) {
    out_mat3x3[0] = in_mat4x4[0];
    out_mat3x3[1] = in_mat4x4[1];
    out_mat3x3[2] = in_mat4x4[2];

    out_mat3x3[3] = in_mat4x4[4];
    out_mat3x3[4] = in_mat4x4[5];
    out_mat3x3[5] = in_mat4x4[6];

    out_mat3x3[6] = in_mat4x4[8];
    out_mat3x3[7] = in_mat4x4[9];
    out_mat3x3[8] = in_mat4x4[10];
}

void mult_matrices(float *output_matrix, float *matrix1, float *matrix2, int matrix_size) {
    int i, j, k;
    float val;

    for (i = 0; i < matrix_size; i++) {
        for (j = 0; j < matrix_size; j++) {
            val = 0.0f;
            for (k = 0; k < matrix_size; k++) {
                val += matrix1[j * matrix_size + k] * matrix2[i + k * matrix_size];
            }
            output_matrix[i + (j * matrix_size)] = val;
        }
    }
}

void mult_matrix_and_scalar(float *output_matrix, float *input_matrix, float scalar, int matrix_cols, int matrix_rows) {
    int i, j;

    for (i = 0; i < matrix_rows; i++) {
        for (j = 0; j < matrix_cols; j++) {
            output_matrix[i * matrix_rows + j] = input_matrix[i * matrix_rows + j] * scalar;
        }
    }
}

float mat2x2_det(float mat[4]) {
    return mat[0] * mat[3] - mat[2] * mat[2];
}

float mat3x3_det(float mat[9]) {
    return mat[0] * mat[4] * mat[8] +
       mat[1] * mat[5] * mat[6] +
       mat[2] * mat[3] * mat[7] -
       mat[2] * mat[4] * mat[6] -
       mat[1] * mat[3] * mat[8] -
       mat[0] * mat[5] * mat[7];
}

void mat4x4_to_mat3x3(float output_mat3x3[9], float input_mat4x4[16]) {
    output_mat3x3[0] = input_mat4x4[0];
    output_mat3x3[1] = input_mat4x4[1];
    output_mat3x3[2] = input_mat4x4[2];
    output_mat3x3[3] = input_mat4x4[4];
    output_mat3x3[4] = input_mat4x4[5];
    output_mat3x3[5] = input_mat4x4[6];
    output_mat3x3[6] = input_mat4x4[8];
    output_mat3x3[7] = input_mat4x4[9];
    output_mat3x3[8] = input_mat4x4[10];
}
void transform_vector(float *output_vector, float *input_matrix, float *input_vector, int vec_matrix_size) {
    int i,j;
    float val;

    for (i = 0; i < vec_matrix_size; i++) {
        val = 0;
        for (j = 0; j < vec_matrix_size; j++) {
            val += input_matrix[i*vec_matrix_size + j] * input_vector[j];
        }
        output_vector[i] = val;
    }
}

void get_rotation_x_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = cos(angle);
    output_matrix[6] = -sin(angle);
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = sin(angle);
    output_matrix[10] = cos(angle);
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_rotation_y_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = cos(angle);
    output_matrix[1] = 0;
    output_matrix[2] = sin(angle);
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = -sin(angle);
    output_matrix[9] = 0;
    output_matrix[10] = cos(angle);
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_rotation_z_matrix4x4(float *output_matrix, float angle) {
    output_matrix[0] = cos(angle);
    output_matrix[1] = -sin(angle);
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = sin(angle);
    output_matrix[5] = cos(angle);
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_translation_matrix4x4v(float *output_matrix, float *v) {
    get_translation_matrix4x4(output_matrix, v[0], v[1], v[2]);
}

void get_translation_matrix4x4(float *output_matrix, float x, float y, float z) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = x;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = y;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = z;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_translation_matrix4x4v_rev(float *output_matrix, float *v) {
    get_translation_matrix4x4(output_matrix, -v[0], -v[1], -v[2]);
}

void get_translation_matrix4x4_rev(float *output_matrix, float x, float y, float z) {
    get_translation_matrix4x4(output_matrix, -x, -y, -z);
}

void get_scale_matrix4x4(float *output_matrix, float scale_x, float scale_y, float scale_z) {
    output_matrix[0] = scale_x;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = scale_y;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = scale_z;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_scale_matrix4x4_rev(float *output_matrix, float scale_x, float scale_y, float scale_z) {
    get_scale_matrix4x4(output_matrix, 1/scale_x, 1/scale_y, 1/scale_z);
}

void get_ortho_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far) {
    output_matrix[0] = 2 / (right - left);
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = -(right + left) / (right - left);

    output_matrix[4] = 0;
    output_matrix[5] = 2 / (top - bottom);
    output_matrix[6] = 0;
    output_matrix[7] = -(top + bottom) / (top - bottom);

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = -2 / (far - near);
    output_matrix[11] = -(far + near) / (far - near);

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void get_frustum_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far) {
    output_matrix[0] = 2 * near / (right - left);
    output_matrix[1] = 0;
    output_matrix[2] = (right + left) / (right - left);
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 2 * near / (top - bottom);
    output_matrix[6] = (top + bottom) / (top - bottom);
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = -(far + near) / (far - near);
    output_matrix[11] = -2 * far * near / (far - near);

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = -1;
    output_matrix[15] = 0;
}

void get_perspective_matrix(float *output_matrix, float fovy, float aspect, float near, float far) {
    float radians = DEG_TO_RAD(fovy);

    output_matrix[0] = 1/(tan(radians/2)*aspect);
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 1/tan(radians/2);
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = (far + near) / (near - far);
    output_matrix[11] = 2 * far * near / (near - far);

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = -1;
    output_matrix[15] = 0;
}

void get_lookat_matrix(float *output_matrix, float eye_x, float eye_y, float eye_z,
                        float center_x, float center_y, float center_z,
                        float up_x, float up_y, float up_z) {
    vector3 f, up, s, s_normalized, u;
    float translation_matrix[16], orientation_matrix[16];
    f.x = center_x - eye_x;
    f.y = center_y - eye_y;
    f.z = center_z - eye_z;

    up.x = up_x;
    up.y = up_y;
    up.z = up_z;

    vector3_normalize(&f);
    vector3_normalize(&up);
    vector3_vector_product(&f, &up, &s);

    s_normalized = s;
    vector3_normalize(&s_normalized);
    vector3_vector_product(&s_normalized, &f, &u);

    orientation_matrix[0] = s.x;
    orientation_matrix[1] = s.y;
    orientation_matrix[2] = s.z;
    orientation_matrix[3] = 0;

    orientation_matrix[4] = u.x;
    orientation_matrix[5] = u.y;
    orientation_matrix[6] = u.z;
    orientation_matrix[7] = 0;

    orientation_matrix[8] = -f.x;
    orientation_matrix[9] = -f.y;
    orientation_matrix[10] = -f.z;
    orientation_matrix[11] = 0;

    orientation_matrix[12] = 0;
    orientation_matrix[13] = 0;
    orientation_matrix[14] = 0;
    orientation_matrix[15] = 1;

    get_translation_matrix4x4(translation_matrix, -eye_x, -eye_y, -eye_z);
    mult_matrices(output_matrix, orientation_matrix, translation_matrix, 4);
}

void get_identity_matrix(float *output_matrix) {
    output_matrix[0] = 1;
    output_matrix[1] = 0;
    output_matrix[2] = 0;
    output_matrix[3] = 0;

    output_matrix[4] = 0;
    output_matrix[5] = 1;
    output_matrix[6] = 0;
    output_matrix[7] = 0;

    output_matrix[8] = 0;
    output_matrix[9] = 0;
    output_matrix[10] = 1;
    output_matrix[11] = 0;

    output_matrix[12] = 0;
    output_matrix[13] = 0;
    output_matrix[14] = 0;
    output_matrix[15] = 1;
}

void transpose_matrix(float *output_matrix, float *input_matrix, int matrix_size) {
    int i, j;

    for (i = 0; i < matrix_size; i++) {
        for (j = 0; j < matrix_size; j++) {
            output_matrix[i*matrix_size + j] = input_matrix[i + j*matrix_size];
        }
    }
}

unsigned char inverse_matrix_2x2(float output_matrix[4], float input_matrix[4]) {
    float det = mat2x2_det(input_matrix);
    if (fabsf(det) < 0.001) {
        memset(output_matrix, 0, sizeof(float) * 4);
        return 0;
    }

    output_matrix[0] =  input_matrix[3] / det;
    output_matrix[1] = -input_matrix[1] / det;

    output_matrix[2] = -input_matrix[2] / det;
    output_matrix[3] =  input_matrix[0] / det;

    return 1;
}

unsigned char inverse_matrix_3x3(float output_matrix[9], float input_matrix[9]) {
    float det = mat3x3_det(input_matrix);
    // for shortening
    float *in = input_matrix;

    int row1 = 0;
    int row2 = 3;
    int row3 = 6;

    int col1 = 0;
    int col2 = 1;
    int col3 = 2;

    //printf("%.2f\n", det);
    if (fabsf(det) < 0.001) {
        memset(output_matrix, 0, sizeof(float) * 9);
        return 0;
    }

    //printf("input_matrix:\n");
    //dump_matrix3x3(input_matrix);
    //printf("\n\n");

    output_matrix[0] =  (in[row2+col2] * in[row3+col3] - in[row2+col3] * in[row3+col2]) / det;
    output_matrix[1] = (in[row1+col3] * in[row3+col2] - in[row1+col2] * in[row3+col3]) / det;
    output_matrix[2] = (in[row1+col2] * in[row2+col3] - in[row1+col3] * in[row2+col2]) / det;

    output_matrix[3] =  (in[row2+col3] * in[row3+col1] - in[row2+col1] * in[row3+col3]) / det;
    output_matrix[4] =  (in[row1+col1] * in[row3+col3] - in[row1+col3] * in[row3+col1]) / det;
    output_matrix[5] =  (in[row1+col3] * in[row2+col1] - in[row1+col1] * in[row2+col3]) / det;

    output_matrix[6] =  (in[row2+col1] * in[row3+col2] - in[row2+col2] * in[row3+col1]) / det;
    output_matrix[7] =  (in[row1+col2] * in[row3+col1] - in[row1+col1] * in[row3+col2]) / det;
    output_matrix[8] =  (in[row1+col1] * in[row2+col2] - in[row1+col2] * in[row2+col1]) / det;

    return 1;
}

unsigned char inverse_matrix_4x4(float output_matrix[16], float input_matrix[16]) {
    double inv[16], det;
    int i;

    inv[0] = input_matrix[5]  * input_matrix[10] * input_matrix[15] -
             input_matrix[5]  * input_matrix[11] * input_matrix[14] -
             input_matrix[9]  * input_matrix[6]  * input_matrix[15] +
             input_matrix[9]  * input_matrix[7]  * input_matrix[14] +
             input_matrix[13] * input_matrix[6]  * input_matrix[11] -
             input_matrix[13] * input_matrix[7]  * input_matrix[10];

    inv[4] = -input_matrix[4]  * input_matrix[10] * input_matrix[15] +
              input_matrix[4]  * input_matrix[11] * input_matrix[14] +
              input_matrix[8]  * input_matrix[6]  * input_matrix[15] -
              input_matrix[8]  * input_matrix[7]  * input_matrix[14] -
              input_matrix[12] * input_matrix[6]  * input_matrix[11] +
              input_matrix[12] * input_matrix[7]  * input_matrix[10];

    inv[8] = input_matrix[4]  * input_matrix[9] * input_matrix[15] -
             input_matrix[4]  * input_matrix[11] * input_matrix[13] -
             input_matrix[8]  * input_matrix[5] * input_matrix[15] +
             input_matrix[8]  * input_matrix[7] * input_matrix[13] +
             input_matrix[12] * input_matrix[5] * input_matrix[11] -
             input_matrix[12] * input_matrix[7] * input_matrix[9];

    inv[12] = -input_matrix[4]  * input_matrix[9] * input_matrix[14] +
               input_matrix[4]  * input_matrix[10] * input_matrix[13] +
               input_matrix[8]  * input_matrix[5] * input_matrix[14] -
               input_matrix[8]  * input_matrix[6] * input_matrix[13] -
               input_matrix[12] * input_matrix[5] * input_matrix[10] +
               input_matrix[12] * input_matrix[6] * input_matrix[9];

    inv[1] = -input_matrix[1]  * input_matrix[10] * input_matrix[15] +
              input_matrix[1]  * input_matrix[11] * input_matrix[14] +
              input_matrix[9]  * input_matrix[2] * input_matrix[15] -
              input_matrix[9]  * input_matrix[3] * input_matrix[14] -
              input_matrix[13] * input_matrix[2] * input_matrix[11] +
              input_matrix[13] * input_matrix[3] * input_matrix[10];

    inv[5] = input_matrix[0]  * input_matrix[10] * input_matrix[15] -
             input_matrix[0]  * input_matrix[11] * input_matrix[14] -
             input_matrix[8]  * input_matrix[2] * input_matrix[15] +
             input_matrix[8]  * input_matrix[3] * input_matrix[14] +
             input_matrix[12] * input_matrix[2] * input_matrix[11] -
             input_matrix[12] * input_matrix[3] * input_matrix[10];

    inv[9] = -input_matrix[0]  * input_matrix[9] * input_matrix[15] +
              input_matrix[0]  * input_matrix[11] * input_matrix[13] +
              input_matrix[8]  * input_matrix[1] * input_matrix[15] -
              input_matrix[8]  * input_matrix[3] * input_matrix[13] -
              input_matrix[12] * input_matrix[1] * input_matrix[11] +
              input_matrix[12] * input_matrix[3] * input_matrix[9];

    inv[13] = input_matrix[0]  * input_matrix[9] * input_matrix[14] -
              input_matrix[0]  * input_matrix[10] * input_matrix[13] -
              input_matrix[8]  * input_matrix[1] * input_matrix[14] +
              input_matrix[8]  * input_matrix[2] * input_matrix[13] +
              input_matrix[12] * input_matrix[1] * input_matrix[10] -
              input_matrix[12] * input_matrix[2] * input_matrix[9];

    inv[2] = input_matrix[1]  * input_matrix[6] * input_matrix[15] -
             input_matrix[1]  * input_matrix[7] * input_matrix[14] -
             input_matrix[5]  * input_matrix[2] * input_matrix[15] +
             input_matrix[5]  * input_matrix[3] * input_matrix[14] +
             input_matrix[13] * input_matrix[2] * input_matrix[7] -
             input_matrix[13] * input_matrix[3] * input_matrix[6];

    inv[6] = -input_matrix[0]  * input_matrix[6] * input_matrix[15] +
              input_matrix[0]  * input_matrix[7] * input_matrix[14] +
              input_matrix[4]  * input_matrix[2] * input_matrix[15] -
              input_matrix[4]  * input_matrix[3] * input_matrix[14] -
              input_matrix[12] * input_matrix[2] * input_matrix[7] +
              input_matrix[12] * input_matrix[3] * input_matrix[6];

    inv[10] = input_matrix[0]  * input_matrix[5] * input_matrix[15] -
              input_matrix[0]  * input_matrix[7] * input_matrix[13] -
              input_matrix[4]  * input_matrix[1] * input_matrix[15] +
              input_matrix[4]  * input_matrix[3] * input_matrix[13] +
              input_matrix[12] * input_matrix[1] * input_matrix[7] -
              input_matrix[12] * input_matrix[3] * input_matrix[5];

    inv[14] = -input_matrix[0]  * input_matrix[5] * input_matrix[14] +
               input_matrix[0]  * input_matrix[6] * input_matrix[13] +
               input_matrix[4]  * input_matrix[1] * input_matrix[14] -
               input_matrix[4]  * input_matrix[2] * input_matrix[13] -
               input_matrix[12] * input_matrix[1] * input_matrix[6] +
               input_matrix[12] * input_matrix[2] * input_matrix[5];

    inv[3] = -input_matrix[1] * input_matrix[6] * input_matrix[11] +
              input_matrix[1] * input_matrix[7] * input_matrix[10] +
              input_matrix[5] * input_matrix[2] * input_matrix[11] -
              input_matrix[5] * input_matrix[3] * input_matrix[10] -
              input_matrix[9] * input_matrix[2] * input_matrix[7] +
              input_matrix[9] * input_matrix[3] * input_matrix[6];

    inv[7] = input_matrix[0] * input_matrix[6] * input_matrix[11] -
             input_matrix[0] * input_matrix[7] * input_matrix[10] -
             input_matrix[4] * input_matrix[2] * input_matrix[11] +
             input_matrix[4] * input_matrix[3] * input_matrix[10] +
             input_matrix[8] * input_matrix[2] * input_matrix[7] -
             input_matrix[8] * input_matrix[3] * input_matrix[6];

    inv[11] = -input_matrix[0] * input_matrix[5] * input_matrix[11] +
               input_matrix[0] * input_matrix[7] * input_matrix[9] +
               input_matrix[4] * input_matrix[1] * input_matrix[11] -
               input_matrix[4] * input_matrix[3] * input_matrix[9] -
               input_matrix[8] * input_matrix[1] * input_matrix[7] +
               input_matrix[8] * input_matrix[3] * input_matrix[5];

    inv[15] = input_matrix[0] * input_matrix[5] * input_matrix[10] -
              input_matrix[0] * input_matrix[6] * input_matrix[9] -
              input_matrix[4] * input_matrix[1] * input_matrix[10] +
              input_matrix[4] * input_matrix[2] * input_matrix[9] +
              input_matrix[8] * input_matrix[1] * input_matrix[6] -
              input_matrix[8] * input_matrix[2] * input_matrix[5];

    det = input_matrix[0] * inv[0] + input_matrix[1] * inv[4] + input_matrix[2] * inv[8] + input_matrix[3] * inv[12];

    if (det == 0) {
        return 0;
    }

    det = 1.0 / det;

    for (i = 0; i < 16; i++) {
        output_matrix[i] = inv[i] * det;
    }

    return 1;
}

char is_point_inside_polygon2(point2 *polygon_vertices, int polygon_vertices_count, point2 *point) {
    int i;
    vector2 v1, v2;
    float result_z;
    char is_point_in_polygon = 1;

    for (i = 0; i < polygon_vertices_count; i++) {
        v1.x = polygon_vertices[(i+1) % polygon_vertices_count].x - polygon_vertices[i].x;
        v1.y = polygon_vertices[(i+1) % polygon_vertices_count].y - polygon_vertices[i].y;

        v2.x = point->x - polygon_vertices[i].x;
        v2.y = point->y - polygon_vertices[i].y;

        result_z = v1.x * v2.y - v1.y * v2.x;
        if (result_z < 0.0f) {
            is_point_in_polygon = 0;
            break;
        }
    }

    return is_point_in_polygon;
}

inline float get_interpolated_value(float Ax, float Ay, float Bx, float By, float Cx) {
    return Ay + (((Cx - Ax) / (Bx - Ax)) * (By - Ay));
}

void get_normal(point3 *p1, point3 *p2, point3 *p3, vector3 *result) {
    vector3 v1, v2;

    v1.x = p1->x - p2->x;
    v1.y = p1->y - p2->y;
    v1.z = p1->z - p2->z;

    v2.x = p1->x - p3->x;
    v2.y = p1->y - p3->y;
    v2.z = p1->z - p3->z;

    result->x = v1.y * v2.z - v1.z * v2.y;
    result->y = v1.z * v2.x - v1.x * v2.z;
    result->z = v1.x * v2.y - v1.y * v2.x;

    vector3_normalize(result);
}

unsigned char check_if_picked_triangle(vector3 *v0, vector3 *v1, vector3 *v2, vector3 *o, vector3 *d) {
    vector3 t1, t2, n, p;
    float front_facing, dist;
    float mat3x3[9], output_matrix[9];
    // uvw_vector - wspolrzedne barycentryczne
    float vector_p[3], uvw_vector[3];

    vector3_sub(v1, v0, &t1);
    vector3_sub(v2, v0, &t2);
    vector3_vector_product(&t1, &t2, &n);

    front_facing = -vector3_scalar_product(&n, d);
    dist = (vector3_scalar_product(&n, o) - vector3_scalar_product(v0, &n)) / front_facing;

    if (dist < 0.0f) {
        return 0;
    }

    // punkt przeciecia prostej z plaszczyczyzna
    t1 = *d;
    vector3_mul(&t1, dist);
    vector3_add(o, &t1, &p);


    mat3x3[0] = v0->x;
    mat3x3[1] = v1->x;
    mat3x3[2] = v2->x;

    mat3x3[3] = v0->y;
    mat3x3[4] = v1->y;
    mat3x3[5] = v2->y;

    mat3x3[6] = v0->z;
    mat3x3[7] = v1->z;
    mat3x3[8] = v2->z;

    inverse_matrix_3x3(output_matrix, mat3x3);
    vector_p[0] = p.x;
    vector_p[1] = p.y;
    vector_p[2] = p.z;

    transform_vector(uvw_vector, output_matrix, vector_p, 3);

    if (uvw_vector[0] < 0.0f || uvw_vector[0] > 1.0f || uvw_vector[1] < 0.0f || uvw_vector[0] + uvw_vector[1] > 1.0f) {
        return 0;
    }

    return 1;
}
