#ifndef __MATHEXT_H__
#define __MATHEXT_H__

#include <math.h>
#include <stdlib.h>

#define DEG_TO_RAD(degrees) (degrees) * M_PI / 180


typedef struct _vector2 {
    float x, y;
} vector2, point2;

typedef struct _vector3 {
    float x, y, z;
} vector3, point3;


/* poniższe dwie funkcje inicjalizują wektor pdanymi wartościami */
vector3 vector3_new(float x, float y, float z);
void vector3_set(vector3 *v, float x, float y, float z);
/* oblicza wartość skalarną wektora */
float vector3_magnitude(vector3 *v);
/* normalizuje podany wektor, czyli zamienia go w wektor tak samo skierowany, ale o wartości skalarnej równej 1 */
void vector3_normalize(vector3 *v);
/* zwraca wektor przeciwny do podanego */
void vector3_reverse(vector3 *v_in, vector3 *v_out);
/* dodawanie wektorów */
void vector3_add(vector3 *v1_in, vector3 *v2_in, vector3 *v_out);
/* odejmowanie wektorów */
void vector3_sub(vector3 *v1_in, vector3 *v2_in, vector3 *v_out);
/* mnożenie wektora przez skalar */
void vector3_mul(vector3 *v, float s);
/* dzielenie wektora przez skalar */
void vector3_div(vector3 *v, float s);
/* zmiana znaku wektora */
void vector3_neg(vector3 *v);

/* Oblicza iloczyn wektorowy (cross product) podanych wektorów. Wynikowy wektor jest prostopadły do obu wektorów,
 * a więc prostopadły do płaszczyzny, w której leżą oba mnożone wektory */
void vector3_vector_product(vector3 *v1_in, vector3 *v2_in, vector3 *v_out);
/* Oblicza iloczyn skalarny (dot product) wektorów. Jest to skalar */
float vector3_scalar_product(vector3 *v1, vector3 *v2);
/* Oblicza skalarny iloczyn mieszany trzech wektorów: u * (v x w). Jest to skalar */
float vector3_triple_scalar_product(vector3 *u, vector3 *v, vector3 *w);

// **************************** UWAGA **********************************
// Macierze (wejściowe i wyjściowe) przechowywane są w orientacji wierszowej:
// wiersz pierwszy: matrix[0] matrix[1] matrix[2] matrix[3]
// wiersz drugi: matrix[4] matrix[5] matrix[6] matrix[7]
// wiersz trzeci: matrix[8] matrix[9] matrix[10] matrix[11]
// wiersz czwarty: matrix[12] matrix[13] matrix[14] matrix[15]
void dump_matrix4x4(float *matrix);
void dump_vector4(float *vector);
void dump_matrix3x3(float *matrix);
void dump_vector3(float *vector);
void dump_coords_3d(float *coords, unsigned long coords_size);
void mat3x3_from_mat4x4(float out_mat3x3[9], float in_mat4x4[16]);
// macierze są mnożone w kolejności matrix1 * matrix2
void mult_matrices(float *output_matrix, float *matrix1, float *matrix2, int matrix_size);
// Mnoży macierz wejściową o podanych wymiarach przez skalar i zwraca wynik w podanej macierzy wyjściowej.
// Macierz wyjściowa jest dokładnie tego samego rozmiaru co macierz wejściowa.
void mult_matrix_and_scalar(float *output_matrix, float *input_matrix, float scalar, int matrix_cols, int matrix_rows);
// macierz i wektor są mnożone w kolejności: input_matrix * input_vector
void transform_vector(float *output_vector, float *input_matrix, float *input_vector, int vec_matrix_size);


// Poniżej kąty są w radianach
void get_rotation_x_matrix4x4(float *output_matrix, float angle);
void get_rotation_y_matrix4x4(float *output_matrix, float angle);
void get_rotation_z_matrix4x4(float *output_matrix, float angle);

void get_translation_matrix4x4v(float *output_matrix, float *v);
void get_translation_matrix4x4(float *output_matrix, float x, float y, float z);
void get_translation_matrix4x4v_rev(float *output_matrix, float *v);
void get_translation_matrix4x4_rev(float *output_matrix, float x, float y, float z);

void get_scale_matrix4x4(float *output_matrix, float scale_x, float scale_y, float scale_z);
void get_scale_matrix4x4_rev(float *output_matrix, float scale_x, float scale_y, float scale_z);

// Transponuje podaną macierz wejściową.
void transpose_matrix(float *output_matrix, float *input_matrix, int matrix_size);
// Odwraca macierz 2x2 i zwraca 1, jeśli macierz odwrotna dla podanej macierzy wejściowej istnieje.
// W przeciwnym razie zwraca 0.
unsigned char inverse_matrix_2x2(float output_matrix[4], float input_matrix[4]);
// Odwraca macierz 3x3 i zwraca 1, jeśli macierz odwrotna dla podanej macierzy wejściowej istnieje.
// W przeciwnym razie zwraca 0.
unsigned char inverse_matrix_3x3(float output_matrix[9], float input_matrix[9]);
// Odwraca macierz 4x4 i zwraca 1, jeśli macierz odwrotna dla podanej macierzy wejściowej istnieje.
// W przeciwnym razie zwraca 0.
unsigned char inverse_matrix_4x4(float output_matrix[16], float input_matrix[16]);

// Poniższe trzy funkcje: get_ortho_matrix, get_frustum_matrix i get_perspective_matrix umieszczają obserwatora
// w początku układu współrzędnych i jest on skierowany w stronę ujemnej półosi OZ. Obserwator jest tak zorientowany
// w przestrzeni, że kierunek "do góry" pokrywa się z kierunkiem dodatniej osi OY.
void get_ortho_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far);
void get_frustum_matrix(float *output_matrix, float left, float right, float bottom, float top, float near, float far);
// fovy - kąt w stopniach widzenia sceny przez obserwatora
void get_perspective_matrix(float *output_matrix, float fovy, float aspect, float near, float far);

void get_lookat_matrix(float *output_matrix, float eye_x, float eye_y, float eye_z,
                        float center_x, float center_y, float center_z,
                        float up_x, float up_y, float up_z);

void get_identity_matrix(float *output_matrix);

// Oblicza wyznacznik macierzy 2x2
float mat2x2_det(float mat[4]);
// Oblicza wyznacznik macierzy 3x3
float mat3x3_det(float mat[9]);
// Zwraca lewą górną część macierzy 4x4 w postaci macierzy 3x3
void mat4x4_to_mat3x3(float output_mat3x3[9], float input_mat4x4[16]);

// **************************** UWAGA **********************************

// wierzchołki wielokąta muszą być umieszczone w kolejności przeciwnej do ruchu wskazówek zegara, wielokąt musi być wypukły
char is_point_inside_polygon2(point2 *polygon_vertices, int polygon_vertices_count, point2 *point);

float get_interpolated_value(float Ax, float Ay, float Bx, float By, float Cx);

// zwraca wektor normalny dla płaszczyzny wyznaczonej przez trzy punkty p1, p2 i p3 (jest to znormalizowany wektor prostopadły do tej płaszczyzny)
void get_normal(point3 *p1, point3 *p2, point3 *p3, vector3 *v_result);

/* Zwraca 1 jeśli odcinek opisany przez punkt początkowy o i wektor kierunkowy d przecina
 * trójkąt opisany wierzchołkami v0, v1, v2.
 * Uwaga! w aktualnej wersji funkcja ta zwraca nieprawidłowe wyniki, należy ją jeszcze sprawdzić i poprawić.
 */
unsigned char check_if_picked_triangle(vector3 *v0, vector3 *v1, vector3 *v2, vector3 *o, vector3 *d);

#endif
